// WTR10-E - read temperature and relative humidity
// 2019-09-14 K.H
//
// ---(LAN)--- MOXA NPort RS-485 serial device server -(RS485 Modbus-RTU)-
//	- WRT10-E - ...

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>

// Using MOXA NPort DE-311
#define HOST	"192.168.3.200"
#define PORT	4001

#ifdef B4800
#define START		8020	// 11*3.5/4800
#elif B2400
#define START		16040	// 11*3.5/2400
#else	// B9600
#define START		4010	// 11*3.5/9600
#endif
#define STOP		START	// 11*3.5/bps

#define MAXRETRY	5
#define TIMEOUT		3

int NPopen(char *, int);
int NPclose(int);
int ReadHoldingRegisters(int, int, int, char *);
unsigned short Crc(unsigned char *, int);
int dump(char*, int);
static void onalrm(int);
void error(char *fmt, ...);
void Usage();

int dflg;
int adr;	// slave address
int fd;
char *cmd;
char buf[BUFSIZ];

int main(int argc, char **argv)
{
  char *s;
  int i;

  cmd = argv[0];
  while (--argc > 0 && (*++argv)[0] == '-')
	for (s = argv[0]+1; *s; s++)
		switch (*s) {
		case 'd':
			dflg++;
			break;
		default:
			Usage();
		}
  if (argc == 0)
	adr = 1;
  else if (argc == 1) {
	adr = atoi(argv[0]);
	if (adr < 1 || 31 < adr)
		error("Illegal slave address %d (must be 1 - 31)", adr);
  }
  if ((fd = NPopen(HOST, PORT)) < 0)
	error("can not open %s", HOST);
  // read humidity and temperature
  if (argc == 0)
	ReadHoldingRegisters(1, 0x0000, 2, NULL);
  else {
	for (i = 0; i < argc; i++) {
		char data[64];

		adr = atoi(argv[i]);
		ReadHoldingRegisters(adr, 0x0000, 2, data);	// 0x03: Read holding registers
		if (i > 0)
			printf(" ");
		printf("%s", data);
	}
	printf("\n");
  }
  NPclose(fd);
  exit(0);
}

static char *err[] = {
  "", "ILLEGAL FUNCTION", "ILLEGAL DATA ADDRESS", "ILLEGAL DATA VALUE",
  "SERVER DEVICE FAILURE", "ACKNOWLEDGE", "SERVER DEVICE BUSY",
  "MEMORY PARITY ERROR", "GATEWAY PATH UNAVAILABLE",
  "GATEWAY TARGET DEVICE FAILED TO RESPOND"
};

int ReadHoldingRegisters(int adr, int reg, int cnt, char *data)
{
  unsigned short crc;
  unsigned char l, h;
  int i, r, c, n;
  short temp, rh;
  char *s;

  *buf = adr & 0xff;	// address
  *(buf + 1) = 0x03;	// Read Holding Registers
  *(unsigned short *)(buf + 2) = htons(reg & 0xffff);
  *(unsigned short *)(buf + 4) = htons(cnt & 0xffff);
  usleep(START);
  crc = Crc((unsigned char *)buf, 6);
  *(unsigned short *)(buf + 6) = htons(crc & 0xffff);
  if (dflg)
	dump(buf, 8);
  if (write(fd, buf, 8) < 0)
	perror("write()");
  *buf = '\0';
  usleep(STOP);
  for (i = 0; i < 9; i++)
	buf[i] = '\0';
  s = buf;
  n = 0;
  alarm(TIMEOUT);
  for (n = 0; n < 9; n += i) {
	if ((i = read(fd, s, 11 - n)) < 0)
		break;
	s += i;
  }
  alarm(0);
  if (n != 9) {
	if (i < 0) {
		if (errno == 4)
			fprintf(stderr, "read() time out.\n");
		else
			fprintf(stderr, "read(): %s\n",  strerror(errno));
	}
	else {
		fprintf(stderr, "read(): data length error (%d byte)\n", n);
		if (dflg)
			dump(buf, 7);
	}
	return -1;
  }
  alarm(0);
  if (dflg)
	dump(buf, n);
  crc =  Crc((unsigned char *)buf, n - 2);
  s = buf + n - 2;
  if (crc != ntohs(*(unsigned short *)s))
	fprintf(stderr, "ReadHoldingRegisters(): receive data CRC error.\n");
  r = buf[1] & 0xff;
  c = buf[2] & 0xff;
  if (r & 0x80) {	// error
	fprintf(stderr, "ReadHoldingRegisters(): ID(%d) reg(0x%04x) %s\n",
		adr, reg, err[c & 0x03]);
	return -1;
  }
  temp = rh = 0.0;
  rh = (buf[3] & 0xff) << 8 | (buf[4] & 0xff);		// relative humidity
  temp = (buf[5] & 0xff) << 8 | (buf[6] & 0xff);	// temperature
  if (data == NULL)
	printf("%.1f %.1f\n", (double)temp * 0.1, (double)rh * 0.1);
  else
	sprintf(data, "%.1f %.1f", (double)temp * 0.1, (double)rh * 0.1);
  return 0;
}

unsigned short Crc(unsigned char *s, int n)
{
  unsigned short crc = 0xffff;
  unsigned short next;
  unsigned short carry;
  unsigned short i;
  unsigned short crcl;

  while (n--) {
	next = (unsigned short)*s;
	crc ^= next;
	for (i = 0; i < 8; i++) {
		carry = crc & 1;
		crc >>= 1;
		if (carry)
			crc ^= 0xA001;
	}
	s++;
  }
  crcl = (crc & 0xff00) >> 8;
  crc <<= 8;
  crc |= crcl;
  return crc;
}

int dump(char *s, int n)
{
  int i;

  for (i = 0; i < n; i++) {
        if (0 < i)
                printf(" ");
        printf("%02X", s[i] & 0xFF);
  }
  printf("\n");
  return 0;
}

static struct sockaddr_in sv_addr;
static struct hostent *shost;
static char srv_addr[80];
static struct sigaction saalrm, osaalrm;

int NPopen(char *host, int port)
{
  int i, r, sofd;

  strcpy(srv_addr, host);	/* IP Address */
  saalrm.sa_handler = onalrm;
  sigemptyset(&saalrm.sa_mask);
  saalrm.sa_flags = 0;
  sigaction(SIGALRM, &saalrm, &osaalrm);
  for (i = 0; i < MAXRETRY; i++) {
	alarm(TIMEOUT);
	if ((sofd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		error("socket() failed", (char *)0);
	if ((shost = gethostbyname(srv_addr)) == NULL)
		error("gethostbyname() failed", (char *)0);
	memset((char *)&sv_addr, 0, sizeof(sv_addr));
	sv_addr.sin_family = AF_INET;
	sv_addr.sin_port = htons(port);
	memcpy((char *)&sv_addr.sin_addr, (char *)shost->h_addr,
		shost->h_length);
	r = connect(sofd, (struct sockaddr *)&sv_addr, sizeof(sv_addr));
	alarm(0);
	if (r == 0)
		break;
	close(sofd);
	sleep(1);
  }
  if (i == MAXRETRY) {
	fprintf(stderr, "NPopen() retry over (%d)\n", i);
	return -1;
  }
  usleep(100000);
  return sofd;
}

int NPclose(int fd)
{
  if (shutdown(fd, 2) < 0)
        error("shutdown() failed", (char *)0);
  if (close(fd) < 0)
	error("cose(%d) failed", fd);
  return 0;
}

static void onalrm(int i)
{
  saalrm.sa_handler = onalrm;
  sigemptyset(&saalrm.sa_mask);
  saalrm.sa_flags = 0;
  sigaction(SIGALRM, &saalrm, &osaalrm);
}

void error(char *fmt, ...)
{
  va_list argp;

  va_start(argp, fmt);
  vfprintf(stderr, fmt, argp);
  va_end(argp);
  fprintf(stderr, "\n");
  exit(1);
}

void Usage()
{
  char *s[] = {
"Comamnd: %s - Read temperature & humidity from WTR10-E",
"Usage: %s [address ..]",
"This program displays relative humidity in %% and temperature in celsius",
"read from the WTR10-E RS485 Modbus-RTU sensor.",
"If address is omitted, 1 is assumed.",
  NULL
  };
  int i;

  for (i = 0; s[i] != NULL; i++) {
	fprintf(stderr, s[i], cmd);
	fprintf(stderr, "\n");
  }
  exit(1);
}
